Path: ersys!pagesat.net!dfw!convex!cs.utexas.edu!swrinde!ringer.cs.utsa.edu!runner!mlevis
From: mlevis@runner.jpl.utsa.edu (Mike Levis - OS/2 Warp II)
Newsgroups: alt.lang.asm
Subject: Re: Efficient asm code
Message-ID: <372rq5$lg7@ringer.cs.utsa.edu>
Date: 7 Oct 1994 07:02:29 GMT
References: <DCMZkS9cp-JL069yn@teleport.com> <36sghl$r60@ringer.cs.utsa.edu> <36tk4q$2slr@vkhdib01.hda.hydro.com>
Organization: The University of Texas at San Antonio
Lines: 334
NNTP-Posting-Host: runner.jpl.utsa.edu
Originator: mlevis@runner

In article <36tk4q$2slr@vkhdib01.hda.hydro.com> terjem@hda.hydro.com (Terje Mathisen) writes:
>You can do a _lot_ better than this, I wrote a function that can search
>for case-ignore strings at the same speed as a case-dependent.

You are absolutely 100% correct.  (actually, Jeff Owens wanted some
suggestions on how to make his search routine smaller, and Boyer-Moore
is bigger).

>It uses Boyer-Moore to skip multiple bytes forward for each tested
>character, with the skip tables encoding the possible case-conversion
>as well.
>
>It scans through my lower 640K of memory for the word 'Copyright' at more
>than 10MB/s on a 486-33.

Here is my Boyer-Moore "delta1+Fast/Slow".  It was adapted to fit the
parameters given by Jeff Owens (using certain registers for input
and output, and support for partial matches).

MakeDelta1 must be called before calling Search so that it can
create the delta table; MakeDelta1 does not have to be called
again unless the pattern string changes, or you need to change
the case sensitivity option.

After calling MakeDelta1, Search can be called one or more times.

CheckTail is called only by Search; CheckTail is a hack that I
was forced to use to allow partial match support (Search also
contains some hacks to support partial matches, which is why
the ASM code does not resemble the C code comments in some places).

The latest "brute-force" search used ~80 bytes of code while
this Boyer-Moore "delta1+Fast/Slow" uses 520+ bytes of code and
650+ bytes of static data.

--------cut here--------

LARGENUM	=	8000h

		dataseg

aiDelta1	dw	256 dup (?)
iPatLen		dw	?
iPatLen2	dw	?
iPatLenMLen2	dw	?
pchTarget	dw	?
pchTargetEnd	dw	?
pchSaveT	dw	?

macro		DELTA	op, dst, src
		mov	bl, src
		sub	bh, bh
		add	bx, bx
		op	dst, [cs:aiDelta1 + bx]
endm

		codeseg

;================
; input:
;    es:bx = pattern
;    dl    = case
;    direction flag = clear
; ouput:
;    [cs:iPatLen] = length of pattern
; destroyed:
;    ax, dh
proc		MakeDelta1
		push	cx
		push	si
		push	di

	; iPatLen = strlen (pchPattern);
		mov	di, bx
		mov	cx, -1
		sub	al, al
		repne	scasb
		not	cx
		dec	cx
		mov	[cs:iPatLen], cx

	; for (i = 0 ; i <= 255 ; i ++)
	;    aiDelta[i] = iPatLen;
		mov	ax, 256
		xchg	ax, cx
		lea	di, [aiDelta1]
		rep	stosw

	; for (i = 0 ; i <= iPatLen ; i ++) {
	;    aiDelta1[pchPattern[i]] = iPatLen - i - 1;
	;    if (((pchPattern[i] | chCase) - 'a') <= 'z')
	;       aiDelta[pchPattern[i]^chCase] = iPatLen - i - 1; }
	; aiDelta1[pchPattern[iPatLen - 1]] = LARGE
	; if (((pchPattern[i] | chCase) - 'a') <= 'z')
	;    aiDelta[pchPattern[iPatLen - 1]^chCase] = LARGE
		xchg	ax, cx
		mov	di, bx
@@loop1:
		cmp	cx, LARGENUM
		je	@@exit
		loop	@@skip2
		mov	cx, LARGENUM
@@skip2:
		mov	al, [es:di]
		inc	di
		mov	si, ax
		add	si, ax
		mov	[cs:aiDelta1 + si], cx

		mov	dh, al
		or	dh, dl
		sub	dh, 'a'
		cmp	dh, 'z' + 1
		jnc	@@loop1
		xor	al, dl
		mov	si, ax
		add	si, ax
		mov	[cs:aiDelta1 + si], cx
		jmp	@@loop1


@@exit:
		pop	di
		pop	si
		pop	cx
		ret
endp		MakeDelta1


;================
;  Called by Search when Boyer-Moore finds no match, but there
;     are still some characters to check. (Used to check partial match)
proc		CheckTail

		mov	si, [cs:pchTargetEnd]
		dec	si
		DELTA	mov, cx, [si]

		mov	di, [cs:iPatLen]
		add	di, dx
		dec	di

@@loop1:
		DELTA	mov, ax, [es:di]
		cmp	ax, cx
		jne	@@skip1

		push	cx
		push	si
		push	di

@@loop2:
		cmp	di, dx
		je	@@match

		dec	si
		dec	di
		DELTA	mov, cx, [si]
		DELTA	mov, ax, [es:di]
		cmp	ax, cx
		je	@@loop2

		pop	di
		pop	si
		pop	cx

@@skip1:
		dec	di
		cmp	di, dx
		jae	@@loop1

		mov	di, dx
		ret

@@match:
		pop	di
		add	sp, 0004h
		inc	di
		ret
endp		CheckTail



;================
; input:
;     es:bx = pattern start
;     es:di = pattern middle
;     ds:si = buffer to search
;     cx    = remaining length of buffer to search
;     [cs:iPatLen] filled in from MakeDelta1
;     [cs:aiDelta1] filled in from MakeDelta1
; output:
;   if match:
;     cx    = non-zero
;     es:si = match end
;   if no match:
;     cx    = zero
;     si    = undefined
;     di    = bx if no partial match (di != bx if partial match)
; destroys:
;     ax
proc		Search
		push	bx
		push	dx
		mov	dx, bx

	; iPatLen2 = iPatLen + pchPattern - pchPattern2;
	; if (iPatLen2 > iTrgLen)
	;    return 0;
		mov	ax, [cs:iPatLen]
		add	ax, dx
		sub	ax, di
		cmp	ax, cx
		jna	@@cont1
		jmp	@@exit
@@cont1:
		mov	[cs:iPatLen2], ax

	; pchT = pchTarget + iPatLen2 - 1;
		mov	[cs:pchTarget], si
		mov	[cs:pchTargetEnd], si
		add	[cs:pchTargetEnd], cx
		add	si, ax
		dec	si
		neg	ax
		add	ax, [cs:iPatLen]
		mov	[cs:iPatLenMLen2], ax


@@fast:
	; pchT += aiDelta1[*pchT];
		DELTA	add, si, [si]


	; if (pchT < pchTarget + iTrgLen)
	;    goto fast
		cmp	si, [cs:pchTargetEnd]
		jb	@@fast

	; if (pchT < LARGE)
	;    return 0;
		cmp	si, LARGENUM
		jae	@@skip10
		call	CheckTail
		jmp	@@no_match
@@skip10:

	; pchT = pchT - LARGE - 1;
	; pchSaveT = pchT + 1;
	; pchP = pchPattern + iPatLen - 2;
		sub	si, LARGENUM
		mov	[cs:pchSaveT], si
		dec	si
		mov	di, dx
		add	di, [cs:iPatLen]
		sub	di, 0002h

@@slow:
	; if (pchP == pchPattern + iPatLen - iPatLen2 - 1)
	;    return pchT + 1;
		mov	ax, dx
		add	ax, [cs:iPatLenMLen2]
		dec	ax
		cmp	di, ax
		je	@@match

	; if (aiDelta1[*pchT] == aiDelta1[*pchP]) {
	;    pchT--; pchP--; goto slow; }
		DELTA	mov, cx, [es:di]
		DELTA	mov, ax, [si]
		cmp	ax, cx
		jne	@@skip1
		dec	si
		dec	di
		jmp	@@slow
@@skip1:

	; iTmp = aiDelta1[*pchT];
	; if (iTmp != LARGE && pchT + iTmp > pchSaveT)
	;    pchT += iTmp;
	; else
	;    pchT = pchSaveT + 2;
		cmp	ax, LARGENUM
		je	@@skip2
		add	ax, si
		mov	cx, [cs:pchSaveT]
		cmp	ax, cx
		jbe	@@skip2
		mov	si, ax
		jmp	short @@skip3
@@skip2:
		mov	si, cx
		add	si, 0002h
@@skip3:

	; pchP = pchPattern;
	; iPatLen2 = iPatLen;
		mov	di, dx
		mov	ax, [cs:iPatLen]
		mov	[cs:iPatLen2], ax
		sub	ax, ax
		mov	[cs:iPatLenMLen2], ax

	; if (pchT < pchTarget + iTrgLen)
	;    goto fast
		cmp	si, [cs:pchTargetEnd]
		jnb	@@no_match
		jmp	@@fast

@@no_match:
		sub	cx, cx
		jmp	short @@exit
@@match:
		add	si, [cs:iPatLen]
		sub	si, di
		add	si, dx
		mov	di, dx
		mov	cx, [cs:pchTargetEnd]
		sub	cx, si
@@exit:
		pop	dx
		pop	bx
		ret
endp		Search

--------cut here--------



-- 
======== Mike Levis   mlevis@runner.jpl.utsa.edu =========
::  OS/2 FAQL and unzip = /os2/all/info/faq/faq21e.zip  ::  .--.
::  & /os2/2_x/archiver/unz512x2.exe, ftp-os2.nmsu.edu  :: (OS/2)
::::::::::::::::: OS/2 3.0 "Warp" beta II ::::::::::::::::  ~--~
